#include "initblock_old.h"

/*
 *----------------------------------------------------------------------
 *
 * HostIF_ReleaseAnonPage --
 *
 *      The page with given mpn is unlocked and released. The page
 *      is put into a cache of released pages. Released pages can be 
 *      recycled later by calling HostIF_GetRecycledPage.
 *
 * Results:
 *      PAGE_UNLOCK_NO_ERROR
 *
 *----------------------------------------------------------------------
 */

int
HostIF_ReleaseAnonPage(VMDriver *vm, MPN mpn)
{
   MemTrackEntry *e;

   e = MemTrack_LookupMPN(vm->memtracker, mpn);
   if (e != NULL) {
      ASSERT(e->mpn == mpn);

      HOST_UNLOCK_PFN(vm, mpn);
      e->mpn = 0;

      MemTrack_RecyclePage(vm->memtracker, e);
   }
   return PAGE_UNLOCK_NO_ERROR;
}


/*
 *----------------------------------------------------------------------
 *
 * HostIF_GetRecycledPage --
 *
 *      Returns the VPN of a previously released anonymous page, if
 *      one is available.
 *
 * Results:
 *      VPN of the page if one is available, INVALID_VPN otherwise
 *
 *----------------------------------------------------------------------
 */

VPN
HostIF_GetRecycledPage(VMDriver *vm)
{
   MemTrackEntry *e;

   e = MemTrack_GetRecycledPage(vm->memtracker);
   if (e != NULL) {
      return e->vpn;
   } else {
      return INVALID_VPN;
   }
}


/*
 *-----------------------------------------------------------------------------
 *
 * HostIF_UserToDriverPtr 
 *
 *    Convert a user virtual address into a kernel virtual address to access
 *    the same physical memory address --hpreg
 *
 * Results:
 *    The kernel virtual address on success
 *    NULL on failure
 *
 * Side effects:
 *    XXX Use 1 per-VM storage location. So for example, we leak kernel memory
 *        when we run a SMP VM on a host with more than 850 MB of RAM --hpreg
 *
 *-----------------------------------------------------------------------------
 */

void *
HostIF_UserToDriverPtr(VMDriver *vm, // IN
                       VA uvAddr)    // IN
{
   struct page *page;

   if (HostIFGetUserPage(vm, (void *)uvAddr, &page)) {
      return NULL;
   }

   HostIF_GlobalLock(17);
   if (vm->vmhost->crosspagePagesCount >= MAX_INITBLOCK_CPUS) {
      HostIF_GlobalUnlock(17);
      put_page(page);
      return NULL;
   }
   vm->vmhost->crosspagePages[vm->vmhost->crosspagePagesCount++] = page;
   HostIF_GlobalUnlock(17);

   return (char *)kmap(page) + (uvAddr & (PAGE_SIZE - 1));
}


/*
 *-----------------------------------------------------------------------------
 *
 * __mod_user_asm --
 *
 *    Mark page dirty. Does not check for kernel address or for broken
 *    write protect on i386.
 *
 * Results:
 *    err unmodified on success
 *    err set to errret on failure.
 *
 * Side effects:
 *    None
 *
 *-----------------------------------------------------------------------------
 */

#define __mod_user_asm(addr, err, errret)	\
   asm volatile (				\
      "1:    lock; orb $0,%1\n"			\
      "2:\n"					\
      ".section .fixup,\"ax\"\n"		\
      "3:    movl %2,%0\n"			\
      "      jmp 2b\n"				\
      ".previous\n"				\
      ".section __ex_table,\"a\"\n"		\
      "      .align 4\n"			\
      "      .long 1b,3b\n"			\
      ".previous"				\
      : "=r"(err)				\
      : "m"(__m(addr)), "i"(errret), "0"(err))


/*
 *-----------------------------------------------------------------------------
 *
 * mod_user_check --
 *
 *    Mark page dirty.
 *
 * Results:
 *    0 on success
 *    -EFAULT on failure - address not mapped, or read-only, or not user address
 *
 * Side effects:
 *    None
 *
 *-----------------------------------------------------------------------------
 */

#define mod_user_check(addr)				\
({							\
   long __mu_err = -EFAULT;				\
   void *__mu_addr = (addr);				\
   if (access_ok(VERIFY_WRITE, __mu_addr, 1)) {		\
      __mu_err = 0;					\
      __mod_user_asm(__mu_addr, __mu_err, -EFAULT);	\
   }							\
   __mu_err;						\
})


/*
 *-----------------------------------------------------------------------------
 *
 * HostIF_MarkPageDirty --
 *
 *    Mark page dirty.
 *
 * Results:
 *    0 on success
 *    != 0 on failure.
 *
 * Side effects:
 *    Complain loudly if global VM lock is held.
 *
 *-----------------------------------------------------------------------------
 */

int
HostIF_MarkPageDirty(VMDriver* vm, void *addr) // IN/OUT
{
#ifdef VM_X86_64
   /* XXX not implemented on x86-64. */
   return 0;
#elif defined(CONFIG_X86_4G) || defined(CONFIG_X86_UACCESS_INDIRECT)
   HostIFGetUserPage(vm, addr, NULL);
   return 0;
#else
   return mod_user_check(addr);
#endif
}


/*
 *----------------------------------------------------------------------
 *
 * HostIF_IsMPNLocked
 *
 *      Return entryPtr if MPN is locked into memory
 *
 * Results:
 *      0 if sucessful and non-zero or error.
 *
 * Side effects:
 *      None.
 *----------------------------------------------------------------------
 */

static void *
FindFunc(void *arg, MemTrackEntry *entryPtr)
{
   MPN mpn = (MPN)(uintptr_t)arg;
   
   if (entryPtr->mpn == mpn) {
      return entryPtr;
   }
   return NULL;
}

int
HostIF_IsMPNLocked(VMDriver *vm, MPN mpn)
{
  MemTrackEntry *entry;

  entry = MemTrack_Scan(vm->memtracker, (void*)(uintptr_t)mpn, FindFunc);
  return (entry != NULL);
}


static void *
CheckFunc(void *arg,               // IN
          MemTrackEntry *entryPtr) // IN
{
   VMDriver *vm = (VMDriver *)arg;
   MPN mpn;
   
   if (entryPtr->mpn == 0) {
      return NULL;
   }

   mpn = HostifVa2Mpn(VPN_2_VA(entryPtr->vpn));
   if (mpn == 0) { 
      /* Unmapped */
      return entryPtr;
   }

   if (entryPtr->mpn != mpn) {
      Warning("CheckFunc vpn 0x%lx mpn 0x%lx and not 0x%lx \n", 
              (unsigned long)entryPtr->vpn, (unsigned long)mpn, (unsigned long)entryPtr->mpn);
      vm->checkFuncFailed = TRUE;
   }
    
   return NULL;
}


/*
 *-----------------------------------------------------------------------------
 *
 * HostIF_CheckMemory
 *
 *      Make sure memory locking is OK.
 *
 *      This function runs as part of the application process, as such
 *      HostifVa2Mpn is fine
 *
 * Results:
 *      None
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------------------------------
 */

VA
HostIF_CheckMemory(VMDriver *vm) // IN
{
   MemTrackEntry *entryPtr;
   
   vm->checkFuncFailed = FALSE;
   entryPtr = MemTrack_Scan(vm->memtracker, vm, CheckFunc);
   
   if (vm->checkFuncFailed) {
      return 1;
   }

   if (entryPtr) {
      return VPN_2_VA(entryPtr->vpn);
   }

   return 0;
}

/*
 *----------------------------------------------------------------------
 *
 * HostIF_MarkLockedVARangeCleanV45 --
 *
 *     Clear the dirty bit in the HW page tables for [VA, VA+len) if
 *     the page is already locked by the monitor/userlevel.
 *
 * Results:
 *     None.
 *
 * Side effects:
 *     None.
 *
 *----------------------------------------------------------------------
 */

int
HostIF_MarkLockedVARangeCleanV45(const VMDriver *vm, VA va, unsigned len)
{
   struct mm_struct *mm = current->mm;
   VA end = va + len;

   if (vm->vmhost->lockedPages == NULL) {
      return 0;
   }
   if (compat_get_page_table_lock(mm)) {
      spin_lock(compat_get_page_table_lock(mm));
   }
   for (;va < end; va += PAGE_SIZE) {
      MPN mpn;
      pte_t *pte;

      pte = PgtblVa2PTELocked(mm, va);
      if (pte != NULL) {
         struct page *page = PgtblPte2Page(pte);
         /* PgtblPte2Page does pte_present. */
         if (pte_dirty(*pte) && page != pfn_to_page(0)) {
            mpn = page_to_pfn(page);
            if (PhysTrack_Test(vm->vmhost->lockedPages, mpn)) {
               uint32 *p = (uint32 *)pte;
               *p &= ~(PTE_D);
            }
         }
         pte_unmap(pte);
      }
   }
   if (compat_get_page_table_lock(mm)) {
      spin_unlock(compat_get_page_table_lock(mm));
   }
   return 0;
}



#define VM_CHECKSIZE(structure, expected)	\
   if (sizeof(structure) != expected) {		\
      asm volatile(".print \"Unexpected ABI change: Structure " #structure " should take %c0 bytes, but took %c1!\"; .abort" : : "i"(expected), "i"(sizeof(structure))); \
   }

#define prop_hex(key,digits,val)			\
   asm volatile(".section \".properties\"\n\t"		\
		".ascii \"" key " \"\n"			\
		"t1 = %a0\n\t"				\
		".rept %c1\n"				\
		"t2 = (t1 >> (%c1 * 4 - 4)) & 15\n\t"	\
		".if t2 < 10\n\t"			\
		".byte t2 + '0'\n\t"			\
		".else\n\t"				\
		".byte t2 + 'A' - 10\n\t"		\
		".endif\n"				\
		"t1 = t1 << 4\n\t"			\
		".endr\n\t"				\
		".byte 10\n\t"				\
		".previous" : : "i"(val), "i"(digits));

#define prop_entry(key, val)				\
   asm volatile(".section \".properties\"\n\t"		\
		".ascii \"" key " " val "\"\n\t"	\
		".byte 10\n\t"				\
		".previous")

#ifdef VMW_HAVE_COMPILE_H
/* Retrieve UTS_VERSION and UTS_MACHINE if there is such information available */
#include <linux/compile.h>
#endif

static void __attribute__((unused))
properties(void)
{
  asm volatile(".section \".properties\", \"\"; .previous");
  prop_entry("UtsRelease", UTS_RELEASE);
#ifdef UTS_VERSION
  prop_entry("UtsVersion", UTS_VERSION);
#endif
#ifdef UTS_MACHINE
  prop_entry("UtsMachine", UTS_MACHINE);
#endif
#ifdef CONFIG_MODVERSIONS
  prop_entry("ModVersions", "yes");
#else
  prop_entry("ModVersions", "no");
#endif
#ifdef CONFIG_SMP
  prop_entry("SMP", "yes");
#else
  prop_entry("SMP", "no");
#endif
#ifdef VM_X86_64
  prop_hex("PageOffset", 16, PAGE_OFFSET);
#else
  prop_hex("PageOffset", 8, PAGE_OFFSET);
#endif
  prop_entry("Comment", "Extended vmmon module from Petr");
  
  /* Make sure that ABI did not change */
  /* We should check crosspage sizes here too... */
  VM_CHECKSIZE(int32, 4);
  VM_CHECKSIZE(MA, 8);
  VM_CHECKSIZE(PA32, 4);
  VM_CHECKSIZE(PA, 8);
  VM_CHECKSIZE(WSLimitsInfo, 24);
  VM_CHECKSIZE(VMMUnlockPageByMPN, 8);
  VM_CHECKSIZE(VMAPICInfo, 8);
  VM_CHECKSIZE(PerfCtrRegisterArgs, 24);
  VM_CHECKSIZE(CPUIDResult, 8);
  VM_CHECKSIZE(CPUIDRegs, 16);
  VM_CHECKSIZE(VMMPNList, 8);
  VM_CHECKSIZE(VMMPNListVA, 12);
  VM_CHECKSIZE(VARange, 12);
  VM_CHECKSIZE(VARangeV45, 8);
  VM_CHECKSIZE(VMMReadWritePage, 8);
  VM_CHECKSIZE(VMMemInfoArgs, 168);
  VM_CHECKSIZE(VMMemInfoArgsV45, 72);
  VM_CHECKSIZE(VMGetMemInfoArgsV4, 48);
  VM_CHECKSIZE(VMSetMemInfoArgsV4, 32);
  VM_CHECKSIZE(VMX86StatsV4, 16);
  VM_CHECKSIZE(VMMemMgmtInfo, 32);
  VM_CHECKSIZE(VMMemMgmtInfoV45, 36);
  VM_CHECKSIZE(VMMemMgmtInfoV4, 32);
  VM_CHECKSIZE(InitBlock, 140);
  VM_CHECKSIZE(InitBlockG25, 164);
  VM_CHECKSIZE(InitBlock30, 36);
  VM_CHECKSIZE(InitBlock20, 48);
  VM_CHECKSIZE(VMGetVMListResult, 8);
  VM_CHECKSIZE(VMGetVMInfoResult, 108);
  VM_CHECKSIZE(VMSetVMInfoArgs, 104);
  VM_CHECKSIZE(CReg, 4);
  VM_CHECKSIZE(Selector, 2);
  VM_CHECKSIZE(Task64, 104);
  VM_CHECKSIZE(Task, 104);
  VM_CHECKSIZE(Task16, 44);
  VM_CHECKSIZE(FarPtr, 6);
  VM_CHECKSIZE(FarPtr16, 4);
  VM_CHECKSIZE(FarPtr32, 6);
  VM_CHECKSIZE(x86ExcFrame64, 40);
  VM_CHECKSIZE(x86ExcFrame32, 12);
  VM_CHECKSIZE(x86ExcFrame32IL, 24);
  VM_CHECKSIZE(x86ExcFrame16, 6);
  VM_CHECKSIZE(x86ExcFrameV8086, 36);
  VM_CHECKSIZE(x86ExcFrameV8086WithErrorCode, 40);
  VM_CHECKSIZE(x86CallStack32, 8);
  VM_CHECKSIZE(x86CallStack16, 4);
  VM_CHECKSIZE(x86CallGateStack32, 16);
  VM_CHECKSIZE(x86CallGateStack16, 8);
  VM_CHECKSIZE(DebugControlRegister, 4);
  VM_CHECKSIZE(SysenterRegs, 12);
  VM_CHECKSIZE(SysenterState, 16);
  VM_CHECKSIZE(SMMSegState, 12);
  VM_CHECKSIZE(SMMStateSaveMap, 512);
  VM_CHECKSIZE(struct FSAVEState, 108);
  VM_CHECKSIZE(struct FXSAVEState, 512);
  VM_CHECKSIZE(struct FXSAVEState64, 512);
}
